from cbpi.utils.encoder import ComplexEncoder
from aiohttp import web
from cbpi.utils.utils import json_dumps
from cbpi.api import request_mapping
import json
class LogHttpEndpoints:

    def __init__(self,cbpi):
        self.cbpi = cbpi
        self.cbpi.register(self, url_prefix="/log")

    @request_mapping(path="/{name}/zip", method="POST", auth_required=False)
    async def create_zip_names(self, request):
        """
        ---
        description: Zip Log files for sensor
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        produces:
        - application/json
        responses:
            "204":
                description: successful operation. Return "pong" text
        """

        log_name = request.match_info['name']
        data = self.cbpi.log.zip_log_data(log_name)

        return web.json_response(dict(filename=data), dumps=json_dumps)

    @request_mapping(path="/{name}/zip", method="DELETE", auth_required=False)
    async def clear_zip_names(self, request):
        """
        ---
        description: Delete all zip files for sensor
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "204":
                description: successful operation.
        """
        log_name = request.match_info['name']
        self.cbpi.log.clear_zip(log_name)
        return web.Response(status=204)

    @request_mapping(path="/zip/download/{name}", method="GET", auth_required=False)
    async def download_zip(self, request):
        """
        ---
        description: Download a sensor zip file
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Zip File name"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "204":
             description: successful operation.
        """

        response = web.StreamResponse(
            status=200,
            reason='OK',
            headers={'Content-Type': 'application/zip'},
        )
        await response.prepare(request)
        log_name = request.match_info['name']
        with open('./logs/%s.zip' % log_name, 'rb') as file:
            for line in file.readlines():
                await response.write(line)

        await response.write_eof()
        return response

    @request_mapping(path="/{name}/zip", method="GET", auth_required=False)
    async def get_zip_names(self, request):
        """
        ---
        description: Zip available zip file names for sensor
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        produces:
        - application/json
        responses:
            "200":
                description: successful operation.
        """
        log_name = request.match_info['name']
        data = self.cbpi.log.get_all_zip_file_names(log_name)
        return web.json_response(data, dumps=json_dumps)

    @request_mapping(path="/{name}/files", method="GET", auth_required=False)
    async def get_file_names(self, request):
        """
        ---
        description: Available log file names for sensor
        tags:
        - Log
        parameters:

        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        produces:
        - application/json
        responses:
            "200":
                description: successful operation.
        """
        log_name = request.match_info['name']
        
        data = self.cbpi.log.get_logfile_names(log_name)
        return web.json_response(data, dumps=json_dumps)

    @request_mapping(path="/{name}", method="GET", auth_required=False)
    async def get_log(self, request):
        """
        ---
        description: delete log data for sensor
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        produces:
        - application/json
        responses:
            "200":
                description: successful operation.
        """
        log_name = request.match_info['name']
        data = await self.cbpi.log.get_data(log_name)
        return web.json_response(data, dumps=json_dumps)


    @request_mapping(path="/", method="POST", auth_required=False)
    async def get_log2(self, request):
        """
        ---
        description: delete log data for sensor
        tags:
        - Log
        parameters:
        - in: body
          name: body
          description: Sensor Ids
          required: true
          schema:
            type: array
            items:
              type: string
        produces:
        - application/json
        responses:
            "200":
                description: successful operation.
        """
        data = await request.json()
        print(data)
        return web.json_response(await self.cbpi.log.get_data2(data), dumps=json_dumps)


    @request_mapping(path="/{name}", method="DELETE", auth_required=False)
    async def clear_log(self, request):
        """
        ---
        description: Get log data for sensor
        tags:
        - Log
        parameters:
        - name: "name"
          in: "path"
          description: "Sensor ID"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "204":
                description: successful operation.
        """
        log_name = request.match_info['name']
        self.cbpi.log.clear_log(log_name)
        return web.Response(status=204)

    @request_mapping(path="/logs", method="POST", auth_required=False)
    async def get_logs(self, request):
        """
        ---
        description: Get Logs
        tags:
        - Log
        parameters:
        - in: body
          name: body
          description: Sensor Ids
          required: true
          schema:
            type: array
            items:
              type: string
        produces:
        - application/json
        responses:
            "200":
                description: successful operation.
        """
        data = await request.json()
        
        result = await self.cbpi.log.get_data(data)
        print("JSON")
        print(json.dumps(result, cls=ComplexEncoder))
        print("JSON----")
        return web.json_response(result, dumps=json_dumps)



import os

from aiohttp import web
from cbpi.api import *

from cbpi.utils import json_dumps
from voluptuous import Schema


class DashBoardHttpEndpoints:

    def __init__(self, cbpi):
        self.cbpi = cbpi
        self.controller = cbpi.dashboard
        self.cbpi.register(self, "/dashboard", os.path.join(cbpi.config_folder.get_file_path("dashboard"), "widgets"))


    @request_mapping(path="/{id:\d+}/content", auth_required=False)
    async def get_content(self, request):
        """
        ---
        description: Get Dashboard Content
        tags:
        - Dashboard
        parameters:
        - name: "id"
          in: "path"
          description: "Dashboard ID"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "200":
                description: successful operation
        """
        dashboard_id = int(request.match_info['id'])
        return web.json_response(await self.cbpi.dashboard.get_content(dashboard_id), dumps=json_dumps)


    @request_mapping(path="/{id:\d+}/content", method="POST", auth_required=False)
    async def add_content(self, request):
        """
        ---
        description: Add Dashboard Content
        tags:
        - Dashboard
        parameters:
        - name: "id"
          in: "path"
          description: "Dashboard ID"
          required: true
          type: "integer"
          format: "int64"
        - name: body
          in: body
          description: Dashboard Content
          required: true
          schema:
            type: object
            properties:
              elements:
                type: array
              pathes:
                type: array
        responses:
            "200":
                description: successful operation
        """
        data = await request.json()
        dashboard_id = int(request.match_info['id'])
        await self.cbpi.dashboard.add_content(dashboard_id, data)
        print("##### SAVE")
        return web.Response(status=204)

    @request_mapping(path="/{id:\d+}/content", method="DELETE", auth_required=False)
    async def delete_conent(self, request):
        """
        ---
        description: Add Dashboard Content
        tags:
        - Dashboard
        parameters:
        - name: "id"
          in: "path"
          description: "Dashboard ID"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "200":
                description: successful operation
        """
  
        dashboard_id = int(request.match_info['id'])
        await self.cbpi.dashboard.delete_content(dashboard_id)
        return web.Response(status=204)

    @request_mapping(path="/widgets", method="GET", auth_required=False)
    async def get_custom_widgets(self, request):
        """
        ---
        description: Get Custom Widgets
        tags:
        - Dashboard
        responses:
            "200":
                description: successful operation
        """
  
      
        return web.json_response(await self.cbpi.dashboard.get_custom_widgets(), dumps=json_dumps)

    @request_mapping(path="/numbers", method="GET", auth_required=False)
    async def get_dashboard_numbers(self, request):
        """
        ---
        description: Get Dashboard Numbers
        tags:
        - Dashboard
        responses:
            "200":
                description: successful operation
        """
        return web.json_response(await self.cbpi.dashboard.get_dashboard_numbers(), dumps=json_dumps)

    @request_mapping(path="/current", method="GET", auth_required=False)
    async def get_current_dashboard(self, request):
        """
        ---
        description: Get Dashboard Numbers
        tags:
        - Dashboard
        responses:
            "200":
                description: successful operation
        """
        return web.json_response(await self.cbpi.dashboard.get_current_dashboard(), dumps=json_dumps)

    @request_mapping(path="/{id}/current", method="POST", auth_required=False)
    async def set_current_dashboard(self, request):
        """
        ---
        description: Set Current Dashboard Number
        tags:
        - Dashboard
        parameters:
        - name: "id"
          in: "path"
          description: "Dashboard ID"
          required: true
          type: "integer"
          format: "int64"
        responses:
            "200":
                description: successful operation
        """
        dashboard_id = int(request.match_info['id'])
        return web.json_response(await self.cbpi.dashboard.set_current_dashboard(dashboard_id))




import os, threading, time, shutil
from aiohttp import web
import logging
from unittest.mock import MagicMock, patch
import asyncio
import random
import json
from cbpi.api import *
from cbpi.api.config import ConfigType
from cbpi.api.base import CBPiBase

logger = logging.getLogger(__name__)

class ConfigUpdate(CBPiExtension):

    def __init__(self,cbpi):
        self.cbpi = cbpi
        self._task = asyncio.create_task(self.run())


    async def run(self):
        logging.info("Check Config for required changes")

        # check is default steps are config parameters

        TEMP_UNIT = self.cbpi.config.get("TEMP_UNIT", "C")
        default_boil_temp = 99 if TEMP_UNIT == "C" else 212
        default_cool_temp = 20 if TEMP_UNIT == "C" else 68
        boil_temp = self.cbpi.config.get("steps_boil_temp", None)
        cooldown_sensor = self.cbpi.config.get("steps_cooldown_sensor", None)
        cooldown_actor = self.cbpi.config.get("steps_cooldown_actor", None)
        cooldown_temp = self.cbpi.config.get("steps_cooldown_temp", None)
        mashin_step = self.cbpi.config.get("steps_mashin", None)
        mash_step = self.cbpi.config.get("steps_mash", None)
        mashout_step = self.cbpi.config.get("steps_mashout", None)
        boil_step = self.cbpi.config.get("steps_boil", None)
        cooldown_step = self.cbpi.config.get("steps_cooldown", None)
        max_dashboard_number = self.cbpi.config.get("max_dashboard_number", None)
        current_dashboard_number = self.cbpi.config.get("current_dashboard_number", None)
        logfiles = self.cbpi.config.get("CSVLOGFILES", None)
        influxdb = self.cbpi.config.get("INFLUXDB", None)
        influxdbaddr = self.cbpi.config.get("INFLUXDBADDR", None)
        influxdbport = self.cbpi.config.get("INFLUXDBPORT", None)
        influxdbname = self.cbpi.config.get("INFLUXDBNAME", None)
        influxdbuser = self.cbpi.config.get("INFLUXDBUSER", None)
        influxdbpwd = self.cbpi.config.get("INFLUXDBPWD", None)
        influxdbcloud = self.cbpi.config.get("INFLUXDBCLOUD", None)
        mqttupdate = self.cbpi.config.get("MQTTUpdate", None)
        PRESSURE_UNIT = self.cbpi.config.get("PRESSURE_UNIT", None)


        if boil_temp is None:
            logger.info("INIT Boil Temp Setting")
            try:
                await self.cbpi.config.add("steps_boil_temp", default_boil_temp, ConfigType.NUMBER, "Default Boil Temperature for Recipe Creation")
            except:
                logger.warning('Unable to update database')

        if cooldown_sensor is None:
            logger.info("INIT Cooldown Sensor Setting")
            try:
                await self.cbpi.config.add("steps_cooldown_sensor", "", ConfigType.SENSOR, "Alternative Sensor to monitor temperature durring cooldown (if not selected, Kettle Sensor will be used)")
            except:
                logger.warning('Unable to update database')

        if cooldown_actor is None:
            logger.info("INIT Cooldown Actor Setting")
            try:
                await self.cbpi.config.add("steps_cooldown_actor", "", ConfigType.ACTOR, "Actor to trigger cooldown water on and off (default: None)")
            except:
                logger.warning('Unable to update database')

        if cooldown_temp is None:
            logger.info("INIT Cooldown Temp Setting")
            try:
                await self.cbpi.config.add("steps_cooldown_temp", default_cool_temp, ConfigType.NUMBER, "Cooldown temp will send notification when this temeprature is reached")
            except:
                logger.warning('Unable to update database')

        if cooldown_step is None:
            logger.info("INIT Cooldown Step Type")
            try:
                await self.cbpi.config.add("steps_cooldown", "", ConfigType.STEP, "Cooldown step type")
            except:
                logger.warning('Unable to update database')

        if mashin_step is None:
            logger.info("INIT MashIn Step Type")
            try:
                await self.cbpi.config.add("steps_mashin", "", ConfigType.STEP, "MashIn step type")
            except:
                logger.warning('Unable to update database')

        if mash_step is None:
            logger.info("INIT Mash Step Type")
            try:
                await self.cbpi.config.add("steps_mash", "", ConfigType.STEP, "Mash step type")
            except:
                logger.warning('Unable to update database')

        if mashout_step is None:
            logger.info("INIT MashOut Step Type")
            try:
                await self.cbpi.config.add("steps_mashout", "", ConfigType.STEP, "MashOut step type")
            except:
                logger.warning('Unable to update database')

        if boil_step is None:
            logger.info("INIT Boil Step Type")
            try:
                await self.cbpi.config.add("steps_boil", "", ConfigType.STEP, "Boil step type")
            except:
                logger.warning('Unable to update database')

        if max_dashboard_number is None:
            logger.info("INIT Max Dashboard Numbers for multiple dashboards")
            try:
                await self.cbpi.config.add("max_dashboard_number", 4, ConfigType.SELECT, "Max Number of Dashboards",
                                                                                                [{"label": "1", "value": 1},
                                                                                                {"label": "2", "value": 2},
                                                                                                {"label": "3", "value": 3},
                                                                                                {"label": "4", "value": 4},
                                                                                                {"label": "5", "value": 5},
                                                                                                {"label": "6", "value": 6},
                                                                                                {"label": "7", "value": 7},
                                                                                                {"label": "8", "value": 8},
                                                                                                {"label": "9", "value": 9},
                                                                                                {"label": "10", "value": 10}])
            except:
                logger.warning('Unable to update database')

        if current_dashboard_number is None:
            logger.info("INIT Current Dashboard Number")
            try:
                await self.cbpi.config.add("current_dashboard_number", 1, ConfigType.NUMBER, "Number of current Dashboard")
            except:
                logger.warning('Unable to update database')

       ## Check if AtuoMode for Steps is in config
        AutoMode = self.cbpi.config.get("AutoMode", None)
        if AutoMode is None:
            logger.info("INIT AutoMode")
            try:
                await self.cbpi.config.add("AutoMode", "Yes", ConfigType.SELECT, "Use AutoMode in steps", 
                                                                                                [{"label": "Yes", "value": "Yes"},
                                                                                                {"label": "No", "value": "No"}])
            except:
                logger.warning('Unable to update config')

        ## Check if AddMashInStep for Steps is in config
        AddMashIn = self.cbpi.config.get("AddMashInStep", None)
        if AddMashIn is None:
            logger.info("INIT AddMashInStep")
            try:
                await self.cbpi.config.add("AddMashInStep", "Yes", ConfigType.SELECT, "Add MashIn Step automatically if not defined in recipe", 
                                                                                                [{"label": "Yes", "value": "Yes"},
                                                                                                {"label": "No", "value": "No"}])
            except:
                logger.warning('Unable to update config')

        ## Check if Brewfather UserID is in config
        bfuserid = self.cbpi.config.get("brewfather_user_id", None)
        if bfuserid is None:
            logger.info("INIT Brewfather User ID")
            try:
                await self.cbpi.config.add("brewfather_user_id", "", ConfigType.STRING, "Brewfather User ID")
            except:
                logger.warning('Unable to update config')

        ## Check if Brewfather API Key is in config
        bfapikey = self.cbpi.config.get("brewfather_api_key", None)
        if bfapikey is None:
            logger.info("INIT Brewfather API Key")
            try:
                await self.cbpi.config.add("brewfather_api_key", "", ConfigType.STRING, "Brewfather API Key")
            except:
                logger.warning('Unable to update config')

        ## Check if Brewfather API Key is in config
        RecipeCreationPath = self.cbpi.config.get("RECIPE_CREATION_PATH", None)
        if RecipeCreationPath is None:
            logger.info("INIT Recipe Creation Path")
            try:
                await self.cbpi.config.add("RECIPE_CREATION_PATH", "upload", ConfigType.STRING, "API path to creation plugin. Default: upload . CHANGE ONLY IF USING A RECIPE CREATION PLUGIN")
            except:
                logger.warning('Unable to update config')

        ## Check if  Kettle for Boil, Whirlpool and Cooldown is in config
        BoilKettle = self.cbpi.config.get("BoilKettle", None)
        if BoilKettle is None:
            logger.info("INIT BoilKettle")
            try:
                await self.cbpi.config.add("BoilKettle", "", ConfigType.KETTLE, "Define Kettle that is used for Boil, Whirlpool and Cooldown. If not selected, MASH_TUN will be used") 
            except:
                logger.warning('Unable to update config')

       ## Check if CSV logfiles is on config 
        if logfiles is None:
            logger.info("INIT CSV logfiles")
            try:
                await self.cbpi.config.add("CSVLOGFILES", "Yes", ConfigType.SELECT, "Write sensor data to csv logfiles", 
                                                                                                [{"label": "Yes", "value": "Yes"},
                                                                                                {"label": "No", "value": "No"}])
            except:
                logger.warning('Unable to update config')

       ## Check if influxdb is on config 
        if influxdb is None:
            logger.info("INIT Influxdb")
            try:
                await self.cbpi.config.add("INFLUXDB", "No", ConfigType.SELECT, "Write sensor data to influxdb", 
                                                                                                [{"label": "Yes", "value": "Yes"},
                                                                                                {"label": "No", "value": "No"}])
            except:
                logger.warning('Unable to update config')

        ## Check if influxdbaddr is in config
        if influxdbaddr is None:
            logger.info("INIT Influxdbaddr")
            try:
                await self.cbpi.config.add("INFLUXDBADDR", "localhost", ConfigType.STRING, "IP Address of your influxdb server (If INFLUXDBCLOUD set to Yes use URL Address of your influxdb cloud server)")
            except:
                logger.warning('Unable to update config')

        ## Check if influxdbport is in config
        if influxdbport is None:
            logger.info("INIT Influxdbport")
            try:
                await self.cbpi.config.add("INFLUXDBPORT", "8086", ConfigType.STRING, "Port of your influxdb server")
            except:
                logger.warning('Unable to update config')

        ## Check if influxdbname is in config
        if influxdbname is None:
            logger.info("INIT Influxdbname")
            try:
                await self.cbpi.config.add("INFLUXDBNAME", "cbpi4", ConfigType.STRING, "Name of your influxdb database name (If INFLUXDBCLOUD set to Yes use bucket of your influxdb cloud database)")
            except:
                logger.warning('Unable to update config')

        ## Check if influxduser is in config
        if influxdbuser is None:
            logger.info("INIT Influxdbuser")
            try:
                await self.cbpi.config.add("INFLUXDBUSER", " ", ConfigType.STRING, "User name for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use organisation of your influxdb cloud database)")
            except:
                logger.warning('Unable to update config')

        ## Check if influxdpwd is in config
        if influxdbpwd is None:
            logger.info("INIT Influxdbpwd")
            try:
                await self.cbpi.config.add("INFLUXDBPWD", " ", ConfigType.STRING, "Password for your influxdb database (only if required)(If INFLUXDBCLOUD set to Yes use token of your influxdb cloud database)")
            except:
                logger.warning('Unable to update config')

       ## Check if influxdb cloud is on config 
        if influxdbcloud is None:
            logger.info("INIT influxdbcloud")
            try:
                await self.cbpi.config.add("INFLUXDBCLOUD", "No", ConfigType.SELECT, "Write sensor data to influxdb cloud (INFLUXDB must set to Yes)", 
                                                                                                [{"label": "Yes", "value": "Yes"},
                                                                                                {"label": "No", "value": "No"}])
            except:
                logger.warning('Unable to update config')

        if mqttupdate is None:
            logger.info("INIT MQTT update frequency for Kettles and Fermenters")
            try:
                await self.cbpi.config.add("MQTTUpdate", 0, ConfigType.SELECT, "Forced MQTT Update frequency in s for Kettle and Fermenter (no changes in payload required). Restart required after change",
                                                                                                [{"label": "30", "value": 30},
                                                                                                {"label": "60", "value": 60},
                                                                                                {"label": "120", "value": 120},
                                                                                                {"label": "300", "value": 300},
                                                                                                {"label": "Never", "value": 0}])
            except:
                logger.warning('Unable to update database')

       ## Check if PRESSURE_UNIT is in config
        if PRESSURE_UNIT is None:
            logger.info("INIT PRESSURE_UNIT")
            try:
                await self.cbpi.config.add("PRESSURE_UNIT", "kPa", ConfigType.SELECT, "Set unit for pressure", 
                                                                                                [{"label": "kPa", "value": "kPa"},
                                                                                                {"label": "PSI", "value": "PSI"}])
            except:
                logger.warning('Unable to update config')

def setup(cbpi):
    cbpi.plugin.register("ConfigUpdate", ConfigUpdate)
    pass



from abc import abstractmethod
import asyncio
import cbpi
import copy
import json
import yaml
import logging
import os.path
import pathlib
from os import listdir
from os.path import isfile, join
import shortuuid
from cbpi.api.dataclasses import  Fermenter, FermenterStep, Props, Step
from cbpi.controller.basic_controller2 import BasicController
from tabulate import tabulate
import sys, os
from ..api.step import CBPiStep, StepMove, StepResult, StepState, CBPiFermentationStep
          
class FermentationController:

    def __init__(self, cbpi):
        self.update_key = "fermenterupdate"
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)
        self.path = self.cbpi.config_folder.get_file_path("fermenter_data.json")
        self.data = []
        self.types = {}
        self.steptypes = {}
        self.cbpi.app.on_cleanup.append(self.shutdown)

    async def init(self):
        logging.info("INIT Fermentation Controller")
        self.check_fermenter_file()
        await self.load()
        pass

    def check_fermenter_file(self):
        if os.path.exists(self.cbpi.config_folder.get_file_path("fermenter_data.json")) is False:
            logging.info("INIT fermenter_data.json file")
            data = {
                    "data": [
                            ]
                    }
            destfile = self.cbpi.config_folder.get_file_path("fermenter_data.json")
            json.dump(data,open(destfile,'w'),indent=4, sort_keys=True)
        
        pathlib.Path(self.cbpi.config_folder.get_file_path("fermenterrecipes")).mkdir(parents=True, exist_ok=True)

    async def shutdown(self, app=None, fermenterid=None):
        self.save()
        if (fermenterid == None):
            for fermenter in self.data:
                self.logger.info("Shutdown {}".format(fermenter.name))
                for step in fermenter.steps:
                    try:
                        self.logger.info("Stop {}".format(step.name))
                        step.instance.shutdown = True
                        await step.instance.stop()
                    except Exception as e:
                        self.logger.error(e)
        else:
            fermenter = self._find_by_id(fermenterid)
            self.logger.info("Shutdown {}".format(fermenter.name))
            for step in fermenter.steps:
                try:
                    self.logger.info("Stop {}".format(step.name))
                    step.instance.shutdown = True
                    await step.instance.stop()
                except Exception as e:
                    self.logger.error(e)


    async def load(self):
        with open(self.path) as json_file:
            data = json.load(json_file)

            for i in data["data"]:
                self.data.append(self._create(i))
                 
    def _create_step(self, fermenter, item):
        id = item.get("id")
        name = item.get("name")
        props = Props(item.get("props"))
        try:
            endtime = int(item.get("endtime", 0))
        except:
            endtime=0

        status = StepState(item.get("status", "I"))
        if status == StepState.ACTIVE:
            status = StepState("S")
        type = item.get("type")

        try:
            type_cfg = self.steptypes.get(type)
            clazz = type_cfg.get("class")
            instance = clazz(self.cbpi, fermenter, item, props, self._done)
        except Exception as e:
            logging.warning("Failed to create step instance %s - %s"  % (id, e))
            instance = None

        step = FermenterStep(id=id, name=name, fermenter=fermenter, props=props, type=type, status=status, endtime=endtime, instance=instance)
        return step

    def _done(self, step_instance, result, fermenter):
        logging.info(result)
        step_instance.step["status"] = "D"
        self.save()
        if result == StepResult.NEXT:
            asyncio.create_task(self.start(fermenter))

    def _create(self, data):
        try:
            id = data.get("id")
            name = data.get("name")
            sensor = data.get("sensor")
            pressure_sensor = data.get("pressure_sensor")
            heater = data.get("heater")
            cooler = data.get("cooler")
            valve = data.get("valve","") 
            logictype = data.get("type")
            temp = data.get("target_temp")
            pressure = data.get("target_pressure")
            brewname = data.get("brewname")
            description = data.get("description")
            props = Props(data.get("props", {}))
            fermenter = Fermenter(id, name, sensor, pressure_sensor, heater, cooler, valve, brewname, description, props, temp, pressure, logictype)
            fermenter.steps = list(map(lambda item: self._create_step(fermenter, item), data.get("steps", [])))
            self.push_update()
            return fermenter
        except:
            return

        
    def _find_by_id(self, id):
        return next((item for item in self.data if item.id == id), None)

    async def get_all(self):
        return list(map(lambda x: x.to_dict(), self.data))
    
    def get_types(self):
        result = {}
        for key, value in self.types.items():
            result[key] = dict(name=value.get("name"), properties=value.get("properties"), actions=value.get("actions"))
        return result

    def get_steptypes(self):
        result = {}
        for key, value in self.steptypes.items():
            result[key] = dict(name=value.get("name"), properties=value.get("properties"), actions=value.get("actions"))
        return result

    def get_state(self):
        if self.data == []:
            pass
        
        return {"data": list(map(lambda x: x.to_dict(), self.data)), "types":self.get_types(), "steptypes":self.get_steptypes()}

    def get_step_state(self, fermenterid=None):
        if self.data == []:
            pass
        fermentersteps=[]
        steplist=list(map(lambda x: x.to_dict(), self.data))
        for fermenter in steplist:
            if fermenterid == fermenter.get("id"):
                fermentersteps={"id": fermenter.get("id"), "steps": fermenter.get("steps")}
        return fermentersteps

    def get_fermenter_steps(self):
        if self.data == []:
            pass
        fermentersteps=[]
        steplist=list(map(lambda x: x.to_dict(), self.data))
        for fermenter in steplist:
            fermenterstep={"id": fermenter.get("id"), "steps": fermenter.get("steps")}
            fermentersteps.append(fermenterstep)
        return fermentersteps

    async def find_step_by_id(self, id):
        actionstep = None
        for item in self.data:
            step = self._find_step_by_id(item.steps, id)
            if step is not None:
                actionstep=step
        return actionstep


    async def get(self, id: str ):
        return self._find_by_id(id)

    async def create(self, data: Fermenter ):
        data.id = shortuuid.uuid()
        self.data.append(data)
        self.save()
        self.push_update()
        return data

    async def update(self, item: Fermenter ):

        def _update(old_item: Fermenter, item: Fermenter):
            old_item.name = item.name
            old_item.sensor = item.sensor
            old_item.pressure_sensor = item.pressure_sensor
            old_item.heater = item.heater
            old_item.cooler = item.cooler
            old_item.valve = item.valve
            old_item.type = item.type
            old_item.brewname = item.brewname
            old_item.description = item.description
            old_item.props = item.props
            old_item.target_temp = item.target_temp
            old_item.target_pressure = item.target_pressure
            return old_item

        self.data = list(map(lambda old: _update(old, item) if old.id == item.id else old, self.data))
        self.save()
        self.push_update()
        return item

    async def set_target_temp(self, id: str, target_temp):
        try:
            item = self._find_by_id(id)
            logging.info(item.target_temp)
            if item:
                item.target_temp = target_temp
                self.save()
                self.push_update()
        except Exception as e:
            logging.error("Failed to set Target Temp {} {}".format(id, e))

    async def set_target_pressure(self, id: str, target_pressure):
        try:
            item = self._find_by_id(id)
            logging.info(item.target_pressure)
            if item:
                item.target_pressure = target_pressure
                self.save()
                self.push_update()
        except Exception as e:
            logging.error("Failed to set Target Pressure {} {}".format(id, e))

    async def delete(self, id: str ):
        item = self._find_by_id(id)
        self.data = list(filter(lambda item: item.id != id, self.data))
        self.save()
        self.push_update()

    def save(self):
        data = dict(data=list(map(lambda item: item.to_dict(), self.data)))         
        with open(self.path, "w") as file:
            json.dump(data, file, indent=4, sort_keys=True)

    def create_step(self, id, item):
        try:
            stepid = shortuuid.uuid()
            item['id'] = stepid
            status = StepState("I")
            type = item.get("type")
            name = item.get("name")
            endtime = item.get("endtime", 0)            
            props = Props(item.get("props"))
            fermenter = self._find_by_id(id)
            try:
                type_cfg = self.steptypes.get(type)
                clazz = type_cfg.get("class")
                instance = clazz(self.cbpi, fermenter, item, props, self._done)
            except Exception as e:
                logging.warning("Failed to create step instance %s - %s"  % (id, e))
                instance = None
            step = FermenterStep(id=stepid, name=name, fermenter=fermenter, props=props, type=type, status=status, endtime=endtime, instance=instance)

            return step
        except Exception as e:
            self.logger.error(e)

    async def update_step(self, id, item):
        fermenter = self._find_by_id(id)
        stepid = item.get("id")
        props = item.get("props")
        status = StepState("I")
        type = item.get("type")
        endtime = 0
        name = item.get("name")
        props = Props(item.get("props"))

        logging.info("update step")
        try:
            type_cfg = self.steptypes.get(type)
            logging.info(type_cfg)
            clazz = type_cfg.get("class")
            logging.info(clazz)
            instance = clazz(self.cbpi, fermenter, item, props, self._done)
            logging.info(instance)
        except Exception as e:
            logging.warning("Failed to create step instance %s - %s "  % (item.id, e))
            instance = None
        step = FermenterStep(id=stepid, name=name, fermenter=fermenter, props=props, type=type, status=status, endtime=endtime, instance=instance)

        try:
            fermenter.steps = list(map(lambda old: step if old.id == step.id else old, fermenter.steps))
        except Exception as e:
            logging.info(e)

        self.save()

        self.push_update("fermenterstepupdate")

    async def delete_step(self, id, stepid):
        item = self._find_by_id(id)
        # might require later check if step is active
        item.steps = list(filter(lambda item: item.id != stepid, item.steps))
        self.save()
        self.push_update("fermenterstepupdate")
    
    async def clearsteps(self, id):
        item = self._find_by_id(id)
        # might require later check if step is active
        item.steps = []
        item.brewname = ""
        self.push_update()
        self.save()
        self.push_update("fermenterstepupdate")

    async def update_endtime(self, id, stepid, endtime):
        try:
            item = self._find_by_id(id)
            step = self._find_step_by_id(item.steps, stepid)
            step.endtime = int(endtime)
            self.save()
            self.push_update("fermenterstepupdate")          
        except Exception as e:
            self.logger.error(e)

    def _find_by_status(self, data, status):
        return next((item for item in data if item.status == status), None)

    def _find_step_by_id(self, data, id):
        return next((item for item in data if item.id == id), None)

    async def update_endtime(self, id, stepid, endtime):
        try:
            item = self._find_by_id(id)
            step = self._find_step_by_id(item.steps, stepid)
            step.endtime = int(endtime)
            self.save()
            self.push_update("fermenterstepupdate")          
        except Exception as e:
            self.logger.error(e)


    async def start(self, id):
        self.logger.info("Start {}".format(id))
        try:
            item = self._find_by_id(id)

            step = self._find_by_status(item.steps, StepState.ACTIVE)
            if step is not None:
                logging.error("Steps already running")
                return

            step = self._find_by_status(item.steps, StepState.STOP)
            if step is not None:
                endtime = step.endtime
                await step.instance.start()
                logging.info("Restarting step {}".format(step.name))
                if endtime != 0:
                    logging.info("Need to change timer")
                step.status = StepState.ACTIVE             
                self.save()
                self.push_update()
                self.push_update("fermenterstepupdate")
                return                     

            step = self._find_by_status(item.steps, StepState.INITIAL)
            logging.info(step)
            if step is None:
                self.logger.info("No futher step to start")
            else:
                step.instance.endtime = 0 
                await step.instance.start()
                logging.info("Starting step {}".format(step.name))
                step.status = StepState.ACTIVE
                self.save()
                self.push_update()
                self.push_update("fermenterstepupdate")

        except Exception as e:
            self.logger.error(e)

    async def stop(self, id):
        try:
            item = self._find_by_id(id)
            step = self._find_by_status(item.steps, StepState.ACTIVE)
            #logging.info(step)
            #logging.info(step.status)
            if step != None:
                logging.info("CALLING STOP STEP")
                try:
                    await step.instance.stop()
                    step.status = StepState.STOP
                    self.save()
                except Exception as e:
                    logging.error("Failed to stop fermenterstep - Id: %s" % step.id)
                self.push_update()
                self.push_update("fermenterstepupdate")

        except Exception as e:
            self.logger.error(e)

    async def start_logic(self, id):
        try:
            item = self._find_by_id(id)
            logging.info("{} Start Id {} ".format(item.name, id))
            if item.instance is not None and item.instance.running is True:
                logging.warning("{} already running {}".format(item.name, id))
                return 
            if item.type is None:
                logging.warning("{} No Type {}".format(item.name, id))
                return 
            clazz = self.types[item.type]["class"]
            item.instance = clazz(self.cbpi, item.id, item.props)
            
            await item.instance.start()
            item.instance.running = True
            item.instance.task = asyncio.get_event_loop().create_task(item.instance._run())
            
            logging.info("{} started {}".format(item.name, id))
            
        except Exception as e:
            logging.error("{} Cant start {} - {}".format(item.name, id, e))

    async def toggle(self, id):
        
        try:
            item = self._find_by_id(id)
            
            if item.instance is None or item.instance.state == False: 
                await self.start_logic(id)
            else:
                await item.instance.stop()
            self.push_update()
            
        except Exception as e:
            logging.error("Failed to switch on FermenterLogic {} {}".format(id, e))


    async  def next(self, id):
        self.logger.info("Next {} ".format(id))
        try:
            item = self._find_by_id(id)
            logging.info(item)
            step = self._find_by_status(item.steps, StepState.ACTIVE)
            logging.info(step)
            if step is not None:
                if step.instance is not None:
                    step.status = StepState.DONE
                    await step.instance.next()
        
            step = self._find_by_status(item.steps, StepState.STOP)
            logging.info(step)
            if step is not None:
                if step.instance is not None:
                    logging.info(step)
                    step.status = StepState.DONE
                    logging.info(step)
                    self.save()
                    await self.start(id)
            else:
                logging.info("No Step is running")
            self.push_update()
            self.push_update("fermenterstepupdate")
        
        except Exception as e:
            self.logger.error(e)


    async def reset(self, id):
        self.logger.info("Reset")
        try:
            item = self._find_by_id(id)
            for step in item.steps:
                self.logger.info("Stopping Step {} {}".format(step.name, step.id))
                try:
                    await step.instance.stop()
                    await step.instance.reset()
                    step.status = StepState.INITIAL
                    step.endtime = 0
                except Exception as e:
                    self.logger.error(e)
            self.save()
            self.push_update()
            self.push_update("fermenterstepupdate")
            
        except Exception as e:
            self.logger.error(e)

    async def move_step(self, fermenter_id, step_id, direction):
        try:
            fermenter = self._find_by_id(fermenter_id)
            index = next((i for i, item in enumerate(fermenter.steps) if item.id == step_id), None)
            if index == None:
                return 
            if index == 0 and direction == -1:
                return
            if index == len(fermenter.steps)-1 and direction == 1:
                return

            fermenter.steps[index], fermenter.steps[index+direction] = fermenter.steps[index+direction], fermenter.steps[index]
            self.save()
            self.push_update()
            self.push_update("fermenterstepupdate")
        
        except Exception as e:
            self.logger.error(e)

    def push_update(self, key="fermenterupdate"):

        if key == self.update_key:
            self.cbpi.ws.send(dict(topic=key, data=list(map(lambda item: item.to_dict(), self.data))))
            
            for item in self.data:
                self.cbpi.push_update("cbpi/{}/{}".format(self.update_key,item.id), item.to_dict())
            pass
        else:
            fermentersteps=self.get_fermenter_steps()
            self.cbpi.ws.send(dict(topic=key, data=fermentersteps))

            # send mqtt update for active femrentersteps
            for fermenter in fermentersteps:
                for step in fermenter['steps']:
                    if step['status'] == 'A':
                        self.cbpi.push_update("cbpi/{}/{}/{}".format(key,fermenter['id'],step['id']), step)
        
    async def call_action(self, id, action, parameter) -> None:
        logging.info("FermenterStep Controller - call Action {} {}".format(id, action))
        try:
            item = await self.find_step_by_id(id)
            logging.info(item)
            await item.instance.__getattribute__(action)(**parameter)
        except Exception as e:
            logging.error("FermenterStep Controller - Failed to call action on {} {} {}".format(id, action, e))
    
    async def savetobook(self, fermenterid):
        name = shortuuid.uuid()
        path = self.cbpi.config_folder.get_fermenter_recipe_by_id(name)
        fermenter=self._find_by_id(fermenterid)
        try:
            brewname = fermenter.brewname
            description = fermenter.description
            
        except:
            brewname = ""
            description = ""
        self.basic_data={"name": brewname, "description": description}

        try:
            fermentersteps = fermenter.steps
        except:
            fermentersteps = []
        data = dict(basic=self.basic_data, steps=list(map(lambda item: item.to_dict(), fermentersteps)))
        with open(path, "w") as file:
            yaml.dump(data, file)

    async def load_recipe(self, data, fermenterid, name):
        try:
            await self.shutdown(None, fermenterid)
        except: 
            pass
        fermenter = self._find_by_id(fermenterid)

        def add_runtime_data(item):
            item["status"] = "I"
            item["endtime"] = 0
            item["id"] = shortuuid.uuid()
            item["props"]["Sensor"] = fermenter.sensor

        list(map(lambda item: add_runtime_data(item), data.get("steps")))
        try:
            fermenter.description = data['basic'].get("desc")
        except:
            fermenter.description = "No Description"
        if name is not None:
            fermenter.brewname = name
        else:
            try:
                fermenter.brewname = data['basic'].get("name")
            except:
                fermenter.brewname = "Fermentation"
        await self.update(fermenter)
        fermenter.steps=[]
        for item in data.get("steps"):
            fermenter.steps.append(self.create_step(fermenterid, item))
        
        self.save()
        self.push_update("fermenterstepupdate")
        
    async def add_step(self, fermenterid, newstep):
        fermenter = self._find_by_id(fermenterid)
        step = self.create_step(fermenterid, newstep)
        fermenter.steps.append(step)
        self.save()
        self.push_update("fermenterstepupdate")
        return step
        



# -*- coding: utf-8 -*-
from cbpi.api import *
from .mqtt_actor import MQTTActor
from .generic_mqtt_actor import GenericMqttActor
from .tasmota_mqtt_actor import TasmotaMqttActor

def setup(cbpi):
    '''
    This method is called by the server during startup
    Here you need to register your plugins at the server

    :param cbpi: the cbpi core
    :return:
    '''
    if str(cbpi.static_config.get("mqtt", False)).lower() == "true":
        cbpi.plugin.register("MQTTActor", MQTTActor)
        cbpi.plugin.register("MQTT Actor (Generic)", GenericMqttActor)
        cbpi.plugin.register("MQTT Actor (Tasmota)", TasmotaMqttActor)


from cbpi.api.dataclasses import Actor
from cbpi.controller.basic_controller2 import BasicController
import logging
from tabulate import tabulate
class ActorController(BasicController):

    def __init__(self, cbpi):
        super(ActorController, self).__init__(cbpi, Actor,"actor.json")
        self.update_key = "actorupdate"
        self.sorting=True

    async def on(self, id, power=None):
        try:
            item = self.find_by_id(id)
            if power is None:
                logging.info("Power is none")
                if item.power:
                    power = item.power
                else:
                    power = 100
            if item.instance.state is False:
                await item.instance.on(power)
                #await self.push_udpate()
                self.cbpi.ws.send(dict(topic=self.update_key, data=list(map(lambda item: item.to_dict(), self.data))),self.sorting)
                self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict(), True)
            else:
                await self.set_power(id, power)
                
        except Exception as e:
            logging.error("Failed to switch on Actor {} {}".format(id, e))

    async def off(self, id):
        try:
            item = self.find_by_id(id)
            if item.instance.state is True:
                await item.instance.off()
                #await self.push_udpate()
                self.cbpi.ws.send(dict(topic=self.update_key, data=list(map(lambda item: item.to_dict(), self.data))),self.sorting)
                self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to switch on Actor {} {}".format(id, e), True)

    async def toogle(self, id):
        try:
            item = self.find_by_id(id)
            instance = item.get("instance")
            await instance.toggle()
            self.cbpi.ws.send(dict(topic=self.update_key, data=list(map(lambda item: item.to_dict(), self.data))),self.sorting)
            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to toggle Actor {} {}".format(id, e))

    async def set_power(self, id, power):
        try:
            item = self.find_by_id(id)
            await item.instance.set_power(power)
        except Exception as e:
            logging.error("Failed to set power {} {}".format(id, e))

    async def actor_update(self, id, power):
        try:
            item = self.find_by_id(id)
            item.power = round(power)
            #await self.push_udpate()
            self.cbpi.ws.send(dict(topic=self.update_key, data=list(map(lambda item: item.to_dict(), self.data))),self.sorting)
            self.cbpi.push_update("cbpi/actorupdate/{}".format(id), item.to_dict())
        except Exception as e:
            logging.error("Failed to update Actor {} {}".format(id, e))



import asyncio
from cbpi.api.dataclasses import NotificationType
import logging
import shortuuid
class NotificationController:

    def __init__(self, cbpi):
        '''
        :param cbpi: craftbeerpi object
        '''
        self.cbpi = cbpi
        self.logger = logging.getLogger(__name__)
        self.callback_cache = {}    
        self.listener = {}
    
    def add_listener(self, method):
        listener_id = shortuuid.uuid()
        self.listener[listener_id] = method
        return listener_id

    def remove_listener(self, listener_id):
        try:
            del self.listener[listener_id] 
        except:
            self.logger.error("Failed to remove listener {}".format(listener_id))

    async def _call_listener(self, title, message, type, action):
        for id, method in self.listener.items():
            print(id, method)
            asyncio.create_task(method(self.cbpi, title, message, type, action ))


    def notify(self, title, message: str, type: NotificationType = NotificationType.INFO, action=[]) -> None:
        '''
        This is a convinience method to send notification to the client
        
        :param key: notification key
        :param message: notification message
        :param type: notification type (info,warning,danger,successs)
        :return: 
        '''
        notifcation_id = shortuuid.uuid()
        
        def prepare_action(item):
            item.id = shortuuid.uuid()
            return item.to_dict()

        actions = list(map(lambda item: prepare_action(item), action))
        self.callback_cache[notifcation_id] = action
        self.cbpi.ws.send(dict(id=notifcation_id, topic="notifiaction", type=type.value, title=title, message=message, action=actions))
        data = dict(type=type.value, title=title, message=message, action=actions)
        self.cbpi.push_update(topic="cbpi/notification", data=data)
        asyncio.create_task(self._call_listener(title, message, type, action))


    def notify_callback(self, notification_id, action_id) -> None:
        try:
            action = next((item for item in self.callback_cache[notification_id]  if item.id == action_id), None)
            if action.method is not None:
                asyncio.create_task(action.method())
            del self.callback_cache[notification_id]
        except Exception as e:
            self.logger.error("Failed to call notificatoin callback")
        


import asyncio
from aiohttp.test_utils import unittest_run_loop
from tests.cbpi_config_fixture import CraftBeerPiTestCase

class StepTestCase(CraftBeerPiTestCase):

    @unittest_run_loop
    async def test_get(self):

        resp = await self.client.request("GET", "/step2")
        print(resp)
        assert resp.status == 200


    @unittest_run_loop
    async def test_crud(self):
        data = {
            "name": "Test",
            "type": "CustomStepCBPi",
            "config": {}
        }

        # Add new step
        resp = await self.client.post(path="/step2/", json=data)
        assert resp.status == 200

        m = await resp.json()
        print("Step", m)
        sensor_id = m["id"]

         # Update step
        resp = await self.client.put(path="/step2/%s" % sensor_id, json=m)
        assert resp.status == 200

        # # Delete step
        resp = await self.client.delete(path="/step2/%s" % sensor_id)
        assert resp.status == 204

    def create_wait_callback(self, topic):
        future = self.cbpi.app.loop.create_future()

        async def test(**kwargs):
            print("GOON")
            future.set_result("OK")
        self.cbpi.bus.register(topic, test, once=True)
        return future

    async def wait(self, future):
        done, pending = await asyncio.wait({future})

        if future in done:
            pass